/*
 *
 * Sample program for use with the Product
 * Licensed Materials  - Property of IBM
 * 5724-I66
 * (c) Copyright IBM Corp.  2008
 *   
 */

package com.ibm.wbit.tel.client.form.command;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.List;

import com.PureEdge.error.UWIException;
import com.PureEdge.xfdl.FormNodeP;
import com.ibm.bpc.clientcore.ClientException;
import com.ibm.bpe.jsf.exception.PropertyNotFoundException;
import com.ibm.bpe.jsf.exception.PropertyPopulationException;
import com.ibm.wbit.tel.client.form.FormUtils;
import com.ibm.wbit.tel.client.form.data.FormDataConversionException;
import com.ibm.wbit.tel.client.jsf.command.BCCreate;
import com.ibm.wbit.tel.client.jsf.command.SubToDoCreateNew;
import com.ibm.wbit.tel.client.jsf.command.ToDoComplete;
import com.ibm.wbit.tel.client.jsf.command.ToDoSave;
import com.ibm.wbit.tel.client.jsf.infrastructure.CommandException;
import com.ibm.wbit.tel.client.jsf.infrastructure.SharedConstants;

/**
 * A facade to execute an action for a form.
 * 
 */
public class FormCommandFacade {

	/**
	 * Command identifier for create case
	 */
	public final static int CREATE_COMMAND = 0;

	/**
	 * Command identifier for save case
	 */
	public final static int SAVE_COMMAND = 1;

	/**
	 * Command identifier for release case
	 */
	public final static int RELEASE_COMMAND = 2;

	/**
	 * Command identifier for complete case
	 */
	public final static int COMPLETE_COMMAND = 3;

	/**
	 * Command identifier for sub todo create case
	 */
	public final static int SUBTASK_CREATE_COMMAND = 4;

	private final static String BC_CREATE_JSP = "BCCreate.jsp";

	private final static String TODOS_UNDER_WORK_JSP = "ToDosUnderWork.jsp";

	private final static String TODOS_OPEN_JSP = "ToDosOpen.jsp";

	private final static String TODOS_COMPLETE_JSP = "ToDoComplete.jsp";

	private final static String INDEX_JSP = "Index.jsp";
	
	private FormNodeP form;
	
	private boolean isInputEqualsOutput=false;
	
	private boolean isMessageEmpty = false;

	public FormCommandFacade(FormNodeP form, boolean isInputEqualsOutput, boolean isMessageEmpty) {
		this.form = form;
		this.isInputEqualsOutput = isInputEqualsOutput;
		this.isMessageEmpty = isMessageEmpty;
	}

	/**
	 * Verifies all signatures in the form and executes the command identified
	 * by parameter <code>type</code>. Possible types are:
	 * <ul>
	 * <li>FormCommandFacade.CREATE_COMMAND</li>
	 * <li>FormCommandFacade.SAVE_COMMAND</li>
	 * <li>FormCommandFacade.RELEASE_COMMAND</li>
	 * <li>FormCommandFacade.COMPLETE_COMMAND</li>
	 * <li>FormCommandFacade.SUBTASK_CREATE_COMMAND</li>
	 * </ul>
	 * 
	 * @param type
	 *            command type
	 * @return next page
	 * @throws ClientException
	 * @throws UWIException
	 * @throws FormDataConversionException 
	 * @throws FormDataConversionException 
	 */
	public String execute(int type) throws ClientException, UWIException, FormDataConversionException {
		// Verify the signatures in the form: if there is a problem, an
		// UWIException is thrown.
		form.verifyAllSignatures(true);

		// If signatures are valid, call the appropriate command.
		String returnPage = FormUtils.createFacesServletURL() + INDEX_JSP;
	
		try{
			if (type == CREATE_COMMAND) {
				returnPage = createTask();
			} else if (type == SAVE_COMMAND) {
				returnPage = saveTask();
			} else if (type == RELEASE_COMMAND) {
				returnPage = releaseTask();
			} else if (type == COMPLETE_COMMAND) {
				returnPage = completeTask();
			} else if (type == SUBTASK_CREATE_COMMAND) {
				returnPage = createSubtask();
			} else {
				throw new IllegalArgumentException();
			}
		}catch(FormDataConversionException e){
			addFormToException(e);
			throw e;
		} catch (CommandException e){
			Throwable cause = e.getCause();			
			if(cause instanceof PropertyNotFoundException || cause instanceof PropertyPopulationException){
				FormDataConversionException childException = new FormDataConversionException(cause.getLocalizedMessage(), cause);
				List<FormDataConversionException> errors = new ArrayList<FormDataConversionException>();
				errors.add(childException);
				FormDataConversionException formDataConversionException = new FormDataConversionException(errors);
				addFormToException(formDataConversionException);	
				throw formDataConversionException;				
			} else{
				throw e;
			}
		}
		return returnPage;
	}
	
	/**
	 * This method will add the content of the form to the FormDataConvresionException.
	 * 
	 * @param formDataConversionException
	 * 		The FormDataException to add the form to. 
	 */
	private void addFormToException(FormDataConversionException formDataConversionException) {
		try{
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			form.writeForm(baos, null, 0);
			baos.flush();
			String formString =  baos.toString("UTF8");
			formDataConversionException.setFormString(formString);			
		} catch (Exception e1){	}
	}

	private String createSubtask() throws ClientException, UWIException, FormDataConversionException {
		FormCreateCommand createCmd = new FormCreateCommand(new SubToDoCreateNew(), form, isInputEqualsOutput);
		return mapListToJSP(createCmd.execute());
	}

	private String completeTask() throws ClientException, UWIException, FormDataConversionException {
		FormToDoCommand completeCmd = new FormToDoCommand(new ToDoComplete(), form, this.isMessageEmpty);
		return mapListToJSP(completeCmd.execute());
	}

	private String saveTask() throws ClientException, UWIException, FormDataConversionException {
		FormToDoCommand saveCmd = new FormToDoCommand(new ToDoSave(), form, this.isMessageEmpty);
		return mapListToJSP(saveCmd.execute());
	}

	private String createTask() throws UWIException, ClientException, FormDataConversionException {
		FormCreateCommand createCmd = new FormCreateCommand(new BCCreate(), form, isInputEqualsOutput);
		createCmd.execute();
		return FormUtils.createFacesServletURL() + BC_CREATE_JSP;
	}

	private String releaseTask() throws ClientException, UWIException, FormDataConversionException {
		ReleaseCommand releaseCmd = new ReleaseCommand();
		return mapListToJSP(releaseCmd.execute());
	}

	String mapListToJSP(String page) {
		String jsp = INDEX_JSP;
		if (SharedConstants.LIST_TODOS_OPEN.equals(page)) {
			jsp = TODOS_OPEN_JSP;
		} else if (SharedConstants.LIST_TODOS_UNDER_WORK.equals(page)) {
			jsp = TODOS_UNDER_WORK_JSP;
		} else if (SharedConstants.DETAILS_TODO_UNDER_WORK.equals(page)) {
			jsp = TODOS_COMPLETE_JSP;
		}
		return FormUtils.createFacesServletURL() + jsp;
	}
}
