/*
 *
 * Sample program for use with the Product
 * Licensed Materials  - Property of IBM
 * 5724-I66
 * (c) Copyright IBM Corp.  2008
 *   
 */

package com.ibm.wbit.tel.client.form.data;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.PureEdge.StringListHolder;
import com.PureEdge.error.UWIException;
import com.PureEdge.xfdl.FormNodeP;
import com.PureEdge.xfdl.XFDL;
import com.ibm.wbit.tel.client.jsf.infrastructure.FacesUtils;
import com.ibm.wbit.tel.client.jsf.infrastructure.Messages;
import com.ibm.wbit.tel.client.jsf.infrastructure.SharedConstants;

/**
 * This class reads and writes the business object data from/to a form. The data
 * will be stored to a map with xpath keys and string values. The map can be
 * used to populate a <code>DataObject</code>.
 */
public class FormDataObject {

	private FormNodeP root;
	
	private FormDataConverter formDataConverter;
	
	private FormDataXpathUtil formDataXpathUtil;
	
	private Map<Document,Map<String,Node>> xpathCache = new HashMap<Document, Map<String,Node>>();
	
	private Map<FormNodeP, Document> instanceDocumentMap = new HashMap<FormNodeP,Document>();
	
	private List<String> formArrayTypes;
	
	private Map<String,String> formsDataTypes;
	
	private String prefix;
	
	@SuppressWarnings("unchecked")
	public FormDataObject(FormNodeP form){
		this.root = form;
		
		Object bean = FacesUtils.getManagedBeanInstance("formsArrayTypes", ArrayList.class);
		if(bean != null){
			this.formArrayTypes = (List<String>)bean;
		}
		
		this.formsDataTypes = (Map<String,String>) FacesUtils.getManagedBeanInstance("formsDataTypes", HashMap.class);
	}

	/**
	 * Read the data from the form and store the information in a map. If no data could be read an exception will be thrown.
	 * 
	 * @param prefix
	 * 				The prefix of the message
	 * 
	 * @throws UWIException
	 * 				If something goes wrong inside the lotus form server api this exception will be thrown.
	 * 
	 * @throws FormDataConversionException 
	 * 			If the data entered into the lotus form does not match the expected format, this exception will be thrown.
	 */
	public HashMap getValues(String prefix) throws UWIException, FormDataConversionException  {
		HashMap<String, Object> values = new HashMap<String, Object>();		
		List<FormDataConversionException> errors = new ArrayList<FormDataConversionException>();
		setPrefix(prefix);
		
		Map<String, String> tempValues = new HashMap<String,String>();
		List<FormNodeP>  instanceNodes = getInstanceNodes(root);	
   	  	for(FormNodeP instance : instanceNodes){                                         
   	  		String instanceID = getInstanceID(instance);  
            FormNodeP firstChild = getFirstNamedChild(instance);  
        	StringListHolder namespaceList = new StringListHolder();
    		StringListHolder attributeList = new StringListHolder();
    		firstChild.getAttributeList(namespaceList, attributeList);
            if(firstChild != null && (firstChild.getChildren() != null || (attributeList.value != null && attributeList.value.length > 0))){
                if (instanceID == null) {
	            	tempValues.putAll(browseDown(firstChild, ""));
	            }
	            else {		       
	            	tempValues.putAll(browseDown(firstChild, "/"+instanceID));
	            }
             }
   	  	}
   	  	
   	  	// build final result.       	  	
   	  	for(String xpath : tempValues.keySet()){   	
       	  	if(isPartOfBusinessObject(xpath)){
				try{
					Object value = convertInput(xpath,tempValues.get(xpath));
					if(value!=null){
						values.put(xpath, value);	
					}
				} catch(FormDataConversionException e){
					errors.add(e);
				}		
       	  	}       	  	
   	  	}

		if(!errors.isEmpty()){
			throw new FormDataConversionException(errors);
		}
		
		if(values.isEmpty() && !tempValues.isEmpty()){
			throw new FormDataConversionException(Messages.getString("FORM_ERROR_CONVERSION_NO_DATA", FacesUtils.getLocale()));
		}
		
		return values;
	}

	/**
	 * This method will browse down a node and return a map containing xpath staments
	 * pointing to all children and their values.
	 * 
	 * @param node
	 * 			The node to inspect.
	 * 
	 * @param xpath
	 * 			The xpath pointing to this node.
	 * 
	 * @return
	 * 			A map with xpath statements and values.
	 * 
	 * @throws UWIException 
	 * 	 		This excepton will be throwan if the Server API cannot process the request.
	 */
	private HashMap<String, String> browseDown(FormNodeP node, String xpath) throws UWIException {
		HashMap<String, String> values = new HashMap<String, String>();
		FormNodeP child = node.getChildren();
		
		// read the attributes.
		StringListHolder namespaceList = new StringListHolder();
		StringListHolder attributeList = new StringListHolder();
		node.getAttributeList(namespaceList, attributeList);
		if(attributeList.value != null){
			for(int i =0 ; i< attributeList.value.length; i++){
				String nameSpace= namespaceList.value[i];
				String attributeName = attributeList.value[i];
				String attributeValue = node.getAttribute(nameSpace,attributeName);
				String attributeXpath =  xpath+ "/@" + attributeName;
				values.put(attributeXpath, attributeValue);
			}	
		}
		
		// read the value out of the leaf
		if(child == null){
			String nodeXpath = xpath;
			if(attributeList.value != null && attributeList.value.length > 0){
				nodeXpath = xpath + "/value";
			}
			values.put(nodeXpath, node.getLiteral());
		} 		
		
		// read the values for of the children.
		while (child != null) {
			String newXpath = getRelativeXpath(child, xpath);
			values.putAll(browseDown(child, newXpath));
			child = child.getNext();
		}
		

		return values;
	}

	/**
	 * This method will return a xpath statement pointing to the node.  
	 * 
	 * @param node
	 * 			The node whichs xpath to find out.
	 * 
	 * @param parentXpath
	 * 			The xpath pointing to the parent node.
	 * 
	 * @return
	 * 			A new xpath pointing to the node.
	 * 
	 * @throws UWIException
	 * 	  		This excepton will be thrown if the Server API cannot process the request.
	 */
	private String getRelativeXpath(FormNodeP node, String parentXpath) throws UWIException {
		return getRelativeXpath(node, parentXpath, null);
	}
	
	/**
	 * This method will return a xpath statement pointing to the node.  
	 * 
	 * @param node
	 * 			The node whichs xpath to find out.
	 * 
	 * @param parentXpath
	 * 			The xpath pointing to the parent node.
	 * 
	 * @param instanceID
	 * 			The id of the instance
	 * 
	 * @return
	 * 			A new xpath pointing to the node.
	 * 
	 * @throws UWIException
	 * 	  		This excepton will be thrown if the Server API cannot process the request.
	 */
	private String getRelativeXpath(FormNodeP node, String parentXpath, String instanceID) throws UWIException {
		String newXpath = parentXpath + "/" + node.getLocalName();
		Integer arrayIndex = getArrayIndex(node, newXpath, instanceID);
		if (arrayIndex != null) {
			newXpath = newXpath + "[" + arrayIndex.intValue() + "]";
		}
		return newXpath;
	}

	/**
	 * This method checks if the the instance data is part of the BO. 
	 * Only instance data that is part of the BO may be passed to the HTM API, otherwise 
	 * an exception will be thrown.
	 * 
	 * @param xpath
	 * 		The xpath of this data node.
	 * 
	 * @param prefix
	 * 		The prefix of the message.
	 * 
	 * @return
	 * 		True if the data instance specified by the xpath an dthe prefix belongs to the Human Task, else no.
	 */
	private boolean isPartOfBusinessObject(String xpath) {
		boolean isPartOfBusinessObject = false;
		String noArrayXPath = getXpathUtil().removeArrayPredicatesFromXpath(xpath);
		String key = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append(noArrayXPath).toString();
		if(getFormsDataTypes().containsKey(key)){
			isPartOfBusinessObject=true;
		}	
		return isPartOfBusinessObject;
	}

	/**
	 * This method will convert the string values deliverd by the server aoi into the  
	 * java data type specified in the business object.
	 * 
	 * @param xpath
	 * 			The xpath of the node.
	 * 
	 * @param literal
	 * 			The value in string form.
	 * 
	 * @param
	 * 		  The refix of this task.
	 * 
	 * @return
	 * 			A correctly typed java object.
	 * 
	 * @throws FormDataConversionException 
	 * 			This exception will be thrown of the value could not be converted.
	 */
	private Object convertInput(String xpath, String literal) throws FormDataConversionException {
		Object value = null;
		String noArrayXPath = getXpathUtil().removeArrayPredicatesFromXpath(xpath);
		String key = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append(noArrayXPath).toString();
		if(getFormsDataTypes().containsKey(key)){
			String type = getFormsDataTypes().get(key);
			value = getConverter().convertInput(literal, type);
		}	
		return value;
	}


	/**
	 * Get the array index of the node. An element will be recognized as an
	 * array participant, if there are more than one element with same name on
	 * same level.
	 * 
	 * @param node
	 * 			The node to get the index of.
	 * 
	 * @param xpath
	 * 			The xpath pointing to the node.
	 * 
	 * @return array index; null, if element is not an array participant.
	 * 
	 * @throws UWIException
	 * 	 		This excepton will be thrown if the Server API cannot process the request.
	 */
	private Integer getArrayIndex(FormNodeP node, String xpath, String instanceID) throws UWIException {
		
		Integer arrayIndexOfNode = null;
		FormNodeP sibling = node.getPrevious();
		int index = 1;
		while (sibling != null){					 
			if(sibling.getLocalName().equals(node.getLocalName())) {
				index++;
			}
			sibling = sibling.getPrevious();
		}
		if(index>1){
			arrayIndexOfNode = new Integer(index);
		}
		if(arrayIndexOfNode == null){
			if(getFormArrayTypes() != null){
				String noArrayXpath = getXpathUtil().removeArrayPredicatesFromXpath(xpath);
				if(instanceID == null){
					noArrayXpath = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append(noArrayXpath).toString();
				}else{
					noArrayXpath = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append("/").append(instanceID).append(noArrayXpath).toString();
				}
				if(getFormArrayTypes().contains(noArrayXpath)){
					arrayIndexOfNode = new Integer(1);
				}				
			}
		}
		return arrayIndexOfNode;
	}
	


	/**
	 * This method sets the values into the form. 
	 * 
	 * @param values
	 * 			A map of keys of type string representing the xpath in the SDO and the value as key.
	 * 
	 * @throws UWIException
	 * 			This excepton will be thrown if the Server API cannot process the request.
	 */
	public void setValues(Map values, String prefix) throws UWIException {
		List<FormNodeP>  instanceNodes = getInstanceNodes(root);
		setPrefix(prefix);

		// build for each instance a DOM object containing the new data instances.
		for(Object key : values.keySet()){
			  String xpath = (String)key;	
			  String firstPart = getXpathUtil().getFirstPart(xpath);
           	  for(FormNodeP instance : instanceNodes){                                         
                  String instanceID = getInstanceID(instance);                  
                  FormNodeP firstChild = getFirstNamedChild(instance);
                  Object value = values.get(key); 
                  if(firstChild != null && value != null){
                      String stringValue = convertOutput(xpath, value);
	                  if (instanceID == null) {
	                	  setValueByXpath(instance,xpath,stringValue);
	                  }
	                  if (firstPart.equals(instanceID)) {
	                      String shortXpath = getXpathUtil().removeFirstPart(xpath);
	                      setValueByXpath(instance,shortXpath,stringValue); 
	                  }
                  }
              }			
		}
		
		// now add missing elements in parents of leaf nodes
		for(Object key : values.keySet()){
			  String xpath = (String)key;						
			  String firstPart = getXpathUtil().getFirstPart(xpath);
         	  for(FormNodeP instance : instanceNodes){                                         
                String instanceID = getInstanceID(instance);                  
                FormNodeP firstChild = getFirstNamedChild(instance);                
                if(firstChild != null){
                      if (instanceID == null) {
	                	  addMissingElements(instance,xpath);
	                  }
	                  if (firstPart.equals(instanceID)) {
	                      String shortXpath = getXpathUtil().removeFirstPart(xpath);
	                      addMissingElements(instance,shortXpath); 
	                  }
                }
            }			
		}
		
		
		// update the instances in xfdl.
		for(FormNodeP instance : instanceDocumentMap.keySet()){	
			updateInstance(instance);
		}
	}
	
	/**
	 * Since the SDO data graph might not be complete, some elements might
	 * be missing, which will lead to a broken UI. Therefore we add the missing
	 * elements by add empty elements. <b>Since we copy the instancedata out of the
	 * form, WE EXPECT EVERY FIRST ARRAY ELEMENT TO BE COMPLETE!<b> Therefore we can 
	 * compare all other elements with the first one.
	 * 
	 * @param children
	 * 			the children that should be completed.
	 * 
	 * @param xpath
	 * 			the xpath of the leaf node
	 * @throws ParserConfigurationException 
	 * @throws UWIException 
	 */
	private void addMissingElements(FormNodeP instance, String xpath) throws UWIException{
		
		String parentXPath = xpath;
		
		while(parentXPath != null && parentXPath.contains("]")){
			parentXPath = getXpathUtil().removeLastPart(parentXPath);
			
			if(parentXPath != null && parentXPath.contains("]"))
			{
				String templateXpath = getXpathUtil().getTemplateXpath(parentXPath);
				try{
					Node parentNode = getNodeForInstanceXpath(instance, parentXPath);
					Node templateNode = getNodeForInstanceXpath(instance, templateXpath);
					addMissingNodesFromTemplateToLeafNode(templateNode, parentNode, templateXpath, getInstanceID(instance));				
				}catch(ParserConfigurationException e){
					
				}
			}
		}
	}

	/**
	 * Add missing elemenets to a node by comparing it to an template node.
	 * This method just perform a flat copy.
	 * 
	 * @param templateNode
	 * 				The node that servers as a template.
	 * @param node
	 * 				The incomplete node.
	 */
	private void addMissingNodesFromTemplateToLeafNode(Node templateNode, Node otherNode, String templateXpath, String instanceID) {
			
		NodeList templateChildren = templateNode.getChildNodes();
		NodeList otherChildren = otherNode.getChildNodes();		
		Document doc = (templateNode).getOwnerDocument();
		
		
		// add missing the elements, be care full with arrays!
		Map<String, Integer> nodeNameCounter = new HashMap<String,Integer>();
		for(int i=0; i<templateChildren.getLength(); i++){
			//get the node name and position
			Node templateChild = templateChildren.item(i);
			// only copy simple elements
			if(mustBeCopied(templateChild,templateXpath, instanceID)){
				String templateNodeName = templateChild.getNodeName();
				int templateNodeNameCount = 1;
				if(nodeNameCounter.containsKey(templateNodeName)){
					templateNodeNameCount = nodeNameCounter.get(templateNodeName).intValue();
					templateNodeNameCount++;
				}		
				nodeNameCounter.put(templateNodeName, templateNodeNameCount);
				
				// see if the missing element is there
				int otherNodeNameCount = 0;
				for(int j=0; j < otherChildren.getLength(); j++){
					Node otherChild = otherChildren.item(j);
					// the node matches if it has the same name and is has the same number
					if(templateNodeName.equals(otherChild.getNodeName())){
						otherNodeNameCount++;
						if(otherNodeNameCount == templateNodeNameCount){
							break;
						}					
					}
				}
				if(otherNodeNameCount < templateNodeNameCount){
					Element newElement = doc.createElement(templateNodeName);
					otherNode.appendChild(newElement);
				}
			} 
			
		}		
		
		// add missing attributes
		NamedNodeMap templateAttributes = templateNode.getAttributes();
		NamedNodeMap otherAttributes = otherNode.getAttributes();
		for(int i=0; i<templateAttributes.getLength(); i++){
			String templateAtrributeName = templateAttributes.item(i).getNodeName();			
			if(otherAttributes.getNamedItem(templateAtrributeName) == null){
				((Element) otherNode).setAttribute(templateAtrributeName, "");
			}
		}
	}
	
	/**
	 * This method decides if a node must be copied. A node must be copied, if
	 * 1) it is an elemenent AND
	 * 2) not an array.
	 * @param node
	 * 		The node to check.
	 * @return
	 * 		true if the node must be copied.
	 */
	private boolean mustBeCopied(Node node, String parentXPath, String instanceID){		
		if(node instanceof Element){
			Element element = (Element)node;
			if(element.getElementsByTagName("*").getLength() == 0){			
				// check if it is an array:
				String childXpath = new StringBuilder(parentXPath).append("/").append(element.getNodeName()).toString();
				String noArrayXpath = getXpathUtil().removeArrayPredicatesFromXpath(childXpath);
				noArrayXpath = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append("/").append(instanceID).append(noArrayXpath).toString();
				if(getFormArrayTypes().contains(noArrayXpath)){
					// arrays must not be copied.
					return false;
				}			
				return true;
			}
		}
		return false;
	}

	/**
	 * This method updates the instance data in the Lotus Form.
	 * 
	 * @param instance
	 * 			The instance to update.
	 * 
	 * @throws UWIException
	 * 			This excepton will be thrown if the Server API cannot process the request.
	 */
	private void updateInstance(FormNodeP instance) throws UWIException {
		try{
			Document doc =instanceDocumentMap.get(instance);			
			String instanceID = getInstanceID(instance);
			OutputFormat format = new OutputFormat(doc);
			format.setIndenting(true);
			ByteArrayOutputStream out = new ByteArrayOutputStream(); 
			XMLSerializer serializer = new XMLSerializer(out, format);
			serializer.serialize(doc);				
			byte[] bytes = out.toByteArray();
			out.flush();				
			String modelID = getModelID(instance);
			String nodeRef;
			if(instanceID == null){
				nodeRef = null;					
			} else{
				StringBuffer buffer = new StringBuffer();
				buffer.append("instance('");
				buffer.append(instanceID);
				buffer.append("')");
				nodeRef = buffer.toString();				
			}
			this.root.updateXFormsInstance(modelID, nodeRef, null,null,bytes, XFDL.UFL_XFORMS_UPDATE_REPLACE);
		} catch(IOException ie) {
		   
		}		
	}
	
	
	/**
	 * This method returns the model id of an instance.
	 * 
	 * @param instance
	 * 			The instance to check.
	 * 
	 * @return
	 * 			The id of the instance as string.
	 * 
	 * @throws UWIException 
	 * 	 		This exception will be thrown if the ServerAPI could not perform the operation.
	 */
	private String getModelID(FormNodeP instance) throws UWIException {
		return getInstanceID(instance.getParent());
	}

	/**
	 * This method will set the value of a certain node specified by the xpath. if the node of the path
	 * do not exist they will be created.
	 * 
	 * @param node
	 * 			The instance node.
	 * 
	 * @param xpath
	 * 			The xpath specifieng the node.
	 * 
	 * @param value
	 * 			The value to set.
	 * 
	 * @throws UWIException 
	 * 	 		This exception will be thrown if the ServerAPI could not perform the operation.
	 */
	private void setValueByXpath(FormNodeP instance, String xpath, String value) throws UWIException {
		try{
			Node node = getNodeForInstanceXpath(instance, xpath);
			if(node!=null){
					Document doc = instanceDocumentMap.get(instance);
					if(node.getFirstChild() != null && node.getFirstChild().getNodeType() == Node.TEXT_NODE){
						node.getFirstChild().setNodeValue(value);
					} else {
						Node valueNode = doc.createTextNode(value);
						node.appendChild(valueNode); 
					}
			}	
		}catch(ParserConfigurationException e){
	
		}
	}
	

	
	/**
	 * This method retuns the node specfied by the xpath relative to the passed instance.
	 * @param node
	 * 			The relative node.
	 * 
	 * @param xpath
	 * 			The xpath node.
	 * 
	 * @return
	 * 			The node specified by the xpath.
	 * 
	 * @throws ParserConfigurationException 
	 * 			This exception will be thrown if the xml support of java is not correctly 
	 * 			configured.
	 * 
	 * @throws ParserConfigurationException 
	 * 	 		This exception will be thrown if the xml support is java is not correctly 
	 * 			configured.
	 * 
	 * @throws UWIException 
	 * 			This exception will be thrown if the server api is not installed correctly.
	 */
	private Node getNodeForInstanceXpath(FormNodeP instance, String xpath) throws UWIException, ParserConfigurationException {
		
		Map<String,Node> cache = getCache(instance);
		
		Node resultNode = null;
		// check if the xpath is allready in cache.
		if(cache.containsKey(xpath)){
			return cache.get(xpath);
		} else {
			// else, go one level deeper.
			String parentXpath = getXpathUtil().removeLastPart(xpath);			
			if(parentXpath == null){
				// we have reached the root, find or create the node with the name and return it.
				resultNode = findOrCreateRootElementNode(instance,xpath);
			} else{
				// ok, get the parent node 
				Node parentNode =  getNodeForInstanceXpath(instance, parentXpath);
				resultNode = findOrCreateNode(instance,parentNode,xpath);
			}			
		}
		
		return resultNode;
	}

	/**
	 * This method will find or create the node specified by this xpath as the 
	 * child of the first child of the root node of the document which belongs to the instance.
	 * 
	 * @param instance
	 * 			The instance node.
	 * 
	 * @param xpath
	 * 			The xpath.
	 * @return
	 * 			The found or create node.
	 * 
	 * @throws UWIException 
	 * 	 		This exception will be thrown if the ServerAPI could not perform the operation.
	 * 
	 * @throws ParserConfigurationException 
	 * 			This exception will be thrown if the xml support is java is not correctly 
	 * 			configured
	 */
	private Node findOrCreateRootElementNode(FormNodeP instance, String xpath) throws UWIException, ParserConfigurationException {
		Document doc = instanceDocumentMap.get(instance);
		return findOrCreateNode(instance, doc.getChildNodes().item(0), xpath);
	}

	
	/**
	 * This method will return the the node specified by the last part of the xpath. If the node does not
	 * exist, it will be created.
	 * 
	 * @param node
	 * 			The parent node of the node to return
	 * 
	 * @param xpath
	 * 			The xpath to create.
	 * 
	 * @return
	 * 			The node specified by the xpath.
	 * 
	 * @throws UWIException 
	 * 	 		This exception will be thrown if the ServerAPI could not perform the operation.
	 * 
	 * @throws ParserConfigurationException 
	 * 			This exception will be thrown if the xml support is java is not correctly 
	 * 			configured
	 */
	private Node findOrCreateNode(FormNodeP instance, Node parentNode, String xpath) throws UWIException, ParserConfigurationException {
		Map<String,Node> cache = getCache(instance);
		
		Node childNode = null;
		
		if(parentNode!= null){
			String nodeName = getXpathUtil().getNodeNameOfLastPart(xpath);
			
			if(nodeName != null && !"".equals(nodeName)){		
				Document doc = instanceDocumentMap.get(instance);	

				// check if the node is an attribute. If not create an element, else
				// an attribute
				if(!nodeName.contains("@")){			
					int position = getXpathUtil().getNodePostion(nodeName);
					nodeName = getXpathUtil().getNodeName(nodeName);
					List<Node> children =getChildrenByByTagName(parentNode, nodeName);
					int nodesFound = children.size();
					if(nodesFound >= position ){
						return children.get(position-1);
					}				
	
					// The node does not exist, create new one(s).
					String parentXpath = getXpathUtil().removeLastPart(xpath);			
					while(nodesFound < position){
							// instaed of only creating a new tree we should copy the entire subtree from element 1, which sould exist
							nodesFound++;	
							String newXpath;
							if(nodesFound == 1 && !xpath.endsWith("]")){
								newXpath  = "/" + nodeName;	
							} else{
								newXpath = "/" + nodeName + "[" + nodesFound + "]";
							}						
							if(parentXpath != null){
								newXpath = parentXpath + newXpath;
							}
							// here i should copy the entire sub tree from the template instance... and update the entire cache... that is a lot of work!
							childNode = doc.createElement(nodeName);
							parentNode.appendChild(childNode);
						
							cache.put(newXpath, childNode);					
					}
				} else {
					String attributeName = getXpathUtil().getAttributeName(nodeName);
					childNode = ((Element) parentNode).getAttributeNode(attributeName);					
					if(childNode == null){
						((Element) parentNode).setAttribute(attributeName, "");
						childNode = ((Element) parentNode).getAttributeNode(attributeName);
						// no need to cache, since it is a leaf.
					}					
				}
			} 
		}		
		return childNode;
	}
	
	/**
	 * returns all children with a given name.
	 * 
	 * @return
	 */
	private List<Node> getChildrenByByTagName(Node parentNode, String nodeName){
		List<Node> children = new ArrayList<Node>();
		NodeList nodes = parentNode.getChildNodes();
		for(int i=0; i< nodes.getLength(); i++){
			Node node = nodes.item(i);
			if(node.getNodeName() != null && node.getNodeName().equals(nodeName)){
				children.add(node);
			}
		}
		return children;
	}

	
	/**
	 * Returns the xpath cache for this instance.  The cache has the following structure. For every XForm instance 
	 * in the Lotus Form there will be an own DOM document object. For each of this documents there will be a map 
	 * containing xpath statements and dom nodes.
	 *  
	 * @param instance
	 * 			The instance.
	 * 
	 * @return
	 * 			A map containing the xpath statments and the node of the xml object.
	 * 
	 * @throws ParserConfigurationException 
	 * 			This exception will be thrown if the xml support is java is not correctly 
	 * 			configured.
	 * 
	 * @throws UWIException 
	 * 			This exception will be thrown if the ServerAPI could not perform the operation.
	 */
	private Map<String, Node> getCache(FormNodeP instance) throws ParserConfigurationException, UWIException {

		if(instanceDocumentMap.get(instance) == null){
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			Document doc = factory.newDocumentBuilder().newDocument();
			FormNodeP firstChild = getFirstNamedChild(instance);
			if(firstChild != null){
				String nameOfFirstChild = firstChild.getLocalName();
				Node rootNode = doc.createElement(nameOfFirstChild);			
				doc.appendChild(rootNode);
				String instanceID = getInstanceID(instance);
				Map<String,Node> cache = copyInstanceToDoc(firstChild, doc, rootNode, "", instanceID);
				instanceDocumentMap.put(instance, doc);	
				xpathCache.put(doc, cache);
			}
		}
		
		Document doc = instanceDocumentMap.get(instance);		
		return xpathCache.get(doc);
	}


	/**
	 * This method will copy the entire xform instance structure into an DOM tree.
	 * 
	 * @param formParent
	 * 			The form node to copy
	 * 
	 * @param doc
	 * 			The docoment to copy to nodes too.
	 * 
	 * @param domParent
	 * 			The node whe the childs node will be appened too.
	 * 
	 * @param xpath
	 * 			The xpath statement pointing to the parent node.
	 * 
	 * @return A map conatining relative xpath statemnets an dthe node the are pointing too.
	 * 
	 * @throws UWIException 
	 * 		   This exception will be thrown if the ServerAPI could not perform the operation.
	 */
	private Map<String,Node> copyInstanceToDoc(FormNodeP formParent, Document doc, Node domParent, String xpath, String instanceID) throws UWIException {
		
		Map<String,Node> cache = new HashMap<String,Node>();		
		FormNodeP formChild = formParent.getChildren();			

		// copy the leaf data
		if(formChild == null){				
			Node textNode = doc.createTextNode(formParent.getLiteral());
			domParent.appendChild(textNode);
		}			
		
		// copy the child and browse down.
		while (formChild != null) {
			Node domChild = doc.createElement(formChild.getLocalName());
			domParent.appendChild(domChild);
			String newXpath = getRelativeXpath(formChild, xpath, instanceID);
			cache.put(newXpath, domChild);		
			cache.putAll(copyInstanceToDoc(formChild, doc, domChild,newXpath, instanceID));
			formChild = formChild.getNext();
		}

		// copy the attibutes
		StringListHolder namespaceList = new StringListHolder();
		StringListHolder attributeList = new StringListHolder();
		formParent.getAttributeList(namespaceList, attributeList);	
		for(int i =0 ; i< attributeList.value.length; i++){
			String nameSpace= namespaceList.value[i];
			String attributeName = attributeList.value[i];
			String attributeValue = formParent.getAttribute(nameSpace,attributeName);
			String attributeXpath =  xpath+ "/@" + attributeName;			
			((Element) domParent).setAttribute(attributeName, attributeValue);
			Node attributeNode = ((Element) domParent).getAttributeNode(attributeName);
			cache.put(attributeXpath, attributeNode);			
		}
		return cache;
	}

	/**
	 * This method will return the first child of an instance node.
	 * 
	 * @param instance
	 * 			The instance node.
	 * 
	 * @return
	 * 		  The first child.
	 * 
	 * @throws UWIException 
	 *      	 This exception will be thrown if the ServerAPI could not perform the operation.
	 */
    private FormNodeP getFirstNamedChild(FormNodeP instance) throws UWIException {
		return instance.getChildren();
	}

    /**
     * This method will return the id of the instance node. This is the value of the "id" attribute.
     * if no id is set nullwill be returned.
     * 
     * @param instance
     * 			The node of the form.
     * 
     * @return
     * 		  The name of the instance or null.
     * 
     * @throws UWIException
     * 		  This exception will be thrown if the ServerAPI could not perform the operation.
     */
	private String getInstanceID(FormNodeP instance) throws UWIException {
		return instance.getAttribute(instance.getNamespaceURI(), "id"); // http://www.w3.org/2002/xforms"
	}


    
    
	/**
	 * Search for all nodes with local name "instance".
	 * 
	 * @param node
	 * 			The root node of the Lotus Form.
	 * 
	 * @return map containg a link between instance nodes and leaf nodes
	 * 
	 * @throws UWIException
	 * This exception will be thrown if the ServerAPI could not perform the operation.
	 */
	private List<FormNodeP> getInstanceNodes(FormNodeP node)
			throws UWIException {
		List<FormNodeP> values = new ArrayList<FormNodeP>();
		while(node != null) {
			if ("instance".equals(node.getLocalName())) {
				values.add(node);
			} else {
				values.addAll(getInstanceNodes(node.getChildren()));
			}
			node = node.getNext();
		}
		return values;
	}
	
	
	
	
	/**
	 * This method will convert to java value into the right string representation for Lotus Forms.
	 * 
	 * @param xpath
	 * 			The Xpath of the node.
	 * 
	 * @param value
	 * 			The value to convert
	 * 
	 * @param prefix
	 * 			The prefix specifing the BO to lookup.
	 * 
	 * @return
	 * 		  The correct String representaion of the value.
	 */
	private String convertOutput(String xpath, Object value){
		String stringValue = "ERROR";		
		String noArrayXPath = getXpathUtil().removeArrayPredicatesFromXpath(xpath);
		String key = new StringBuilder(getPrefix()).append(SharedConstants.NAVIGATION_SEPARATOR).append(noArrayXPath).toString();
		if(getFormsDataTypes().containsKey(key)){
			String type = getFormsDataTypes().get(key);
			stringValue = getConverter().convertOutput(value, type);
		} 
		return stringValue;
	}
	

    
    /**
     * Singleton for the FormDataConverter. The method returns the
     * FormDataConverter that must be invoked to convert the data between
     * Java and Lotus Forms.
     * 
     * @return
     * 		The FormDataConverter.
     */
	private FormDataConverter getConverter(){
		if(formDataConverter == null){
			formDataConverter = new FormDataConverter();
		}		
		return formDataConverter;
	}
	
	/**
	 * Singleton method for the FormDataXpathUtil.
	 * 
	 * @return
	 * 		The FormDataXpathUtil.
	 */
	private FormDataXpathUtil getXpathUtil(){
		if(formDataXpathUtil == null){
			formDataXpathUtil = new FormDataXpathUtil();
		}
		return formDataXpathUtil;
	}
	
	/**
	 * Getter for the root node of the Lotus Form.
	 * 
	 * @return
	 * 		The root node of the LotusForm.
	 */
	public FormNodeP getForm() {
		return this.root;
	}
	
	/**
	 * Getter method for the list of xpaths pointing to array elements.
	 * 
	 * @return
	 * 		A list of xpaths pointing to array elements.
	 */
	private List<String> getFormArrayTypes(){
		return formArrayTypes;
	}

	/**
	 * Getter method for the the map which contains the datatypes
	 * for each element of the business object.
	 * 
	 * @return
	 * 		A lmap which contains the datatypes for each element of the business object.
	 */
	private Map<String, String> getFormsDataTypes() {
		return formsDataTypes;
	}
	
	/**
	 * Getter for the prefix. 
	 * 
	 * @return
	 * 		The prefix for this request.
	 */
	private String getPrefix() {
		return prefix;
	}

	/**
	 * Setter for the prefix.
	 * 
	 * @param prefix
	 * 		Thhe prefix for this request.
	 */
	private void setPrefix(String prefix) {
		this.prefix = prefix;
	}

}
