/*
 *
 * Sample program for use with the Product
 * Licensed Materials  - Property of IBM
 * 5724-I66
 * (c) Copyright IBM Corp.  2008
 *   
 */

package com.ibm.wbit.tel.client.form.persistence;

import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;

import com.PureEdge.IFSSingleton;
import com.PureEdge.error.UWIException;
import com.PureEdge.xfdl.FormNodeP;
import com.ibm.wbit.tel.client.jsf.infrastructure.SharedConstants;

/**
 * A service to save/get forms to/from a storage. Currently the forms will
 * be saved to the java.io.temp directory. To change this, the appropriate
 * methods (<code>getForm(..)</code> and <code>saveForm(..)</code>) must
 * be replaced.
 */
public class PersistenceService {

	private final static String FORMS_DIR = "IBMLotusForms";

	private String tgtDirName;

	public PersistenceService() {
		tgtDirName = System.getProperty("java.io.tmpdir") + FORMS_DIR + File.separator;
	}

	/**
	 * Get a persisted version of the form.
	 * 
	 * @param persistenceInfo
	 * @return the persisted form; null, if no persisted form is available
	 * @throws PersistenceException
	 */
	public FormNodeP getForm(PersistenceInfo persistenceInfo) throws PersistenceException {
		FormNodeP form = null;
		try {
			String formLocation = persistenceInfo.getFormLocation();
			if(persistenceInfo.getContainmentContextId() != null)
			{
				String id = persistenceInfo.getContainmentContextId().toString();
	
				// Get the target directory
				String[] segments = formLocation.split("\\/");
				StringBuffer path = createFormDirPath(id, segments);
				path.append(segments[segments.length - 1]);
	
				// Read the xfdl file
				File formFile = new File(path.toString());
				if(formFile.exists()){
					try {
						FileInputStream inputStream = new FileInputStream(formFile);
						form = IFSSingleton.getXFDL().readForm(inputStream, 0);
					} catch (FileNotFoundException e) {
						form = null;
					}
				}
			}	
		} catch (UWIException e) {
			throw new PersistenceException(SharedConstants.NLS_CATALOG, SharedConstants.FAILED_TO_GET_PERSISTED_FORM, null, e);
		}
		return form;
	}

	/**
	 * Save the form.
	 */
	public void saveForm(PersistenceInfo persistenceInfo, FormNodeP form) throws PersistenceException {
		try {
			String formLocation = persistenceInfo.getFormLocation();
			String id = persistenceInfo.getContainmentContextId().toString();

			// Create the target directory
			String[] segments = formLocation.split("\\/");
			StringBuffer path = createFormDirPath(id, segments);
			File dir = new File(path.toString());
			dir.mkdirs();

			// Create the xfdl file
			path.append(segments[segments.length - 1]);
			File formFile = new File(path.toString());
			formFile.createNewFile();
			BufferedWriter writer = new BufferedWriter(new FileWriter(formFile, false));
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			form.writeForm(baos, null, 0);
			baos.flush();
			writer.write(baos.toString());
			writer.close();
		} catch (IOException e) {
			throw new PersistenceException(SharedConstants.NLS_CATALOG, SharedConstants.FAILED_TO_PERSIST_FORM, null, e);
		} catch (UWIException e) {
			throw new PersistenceException(SharedConstants.NLS_CATALOG, SharedConstants.FAILED_TO_PERSIST_FORM, null, e);
		}
	}

	/**
	 * Creates a directory path: <code>id\segment[0]\...\segment[i]</code>.
	 */
	private StringBuffer createFormDirPath(String id, String[] segments) {
		StringBuffer path = new StringBuffer();
		path.append(tgtDirName);
		id = id.replaceAll("\\:", ".");
		path.append(id);
		for (int index = 0; index < segments.length - 1; index++) {
			String segment = segments[index];
			path.append(File.separator);
			path.append(segment);
		}
		return path.append(File.separator);
	}

	/**
	 * This method will be called after the persistence service was used.
	 */
	public void destroy() {
	}
}
